package argparse

import (
	"errors"
	"fmt"
	"os"
	"reflect"
	"strings"
)

type Parser struct {
	args     []*arg
	showHelp bool
	parsed   bool
}

type Option struct {
	Require bool
}

type arg struct {
	sname string
	lname string
	// value item count
	size   int
	value  interface{}
	unique bool
	parsed bool
	opts   *Option
}

func New() *Parser {
	return &Parser{
		args: make([]*arg, 0),
	}
}

func (p *Parser) addArg(a *arg) error {
	for _, v := range p.args {
		if v.sname == a.sname || v.lname == a.lname {
			return errors.New("option name dup")
		}
	}
	p.args = append(p.args, a)

	return nil
}

func (p *Parser) typeVar(i interface{}, short, long string, size int, unique bool, opts *Option) {
	t := reflect.ValueOf(i)
	if t.Kind() != reflect.Ptr {
		panic(errors.New("var type not ptr"))
	}
	if short == "" && long == "" {
		panic(errors.New("short name and long name is empty"))
	}

	a := &arg{
		sname:  short,
		lname:  long,
		value:  i,
		size:   size,
		unique: unique,
		opts:   opts,
	}

	if err := p.addArg(a); err != nil {
		panic(fmt.Errorf("unable to add String: %s\n", err))
	}
}

func (p *Parser) Help(short, long string) {
	p.typeVar(&p.showHelp, short, long, 0, true, nil)
}

func (p *Parser) printHelp() {
	sb := &strings.Builder{}
	sb.WriteString("Usage:\n")
	for _, v := range p.args {
		if _, err := sb.WriteString(fmt.Sprintf("\t%s\t%s\n", v.name(), v.getType())); err != nil {
			panic(err)
		}
	}
	fmt.Println(sb.String())
	// if flag.Lookup("test.v") == nil {
	os.Exit(2)
	// }
}

func (p *Parser) Parse(a []string) error {
	copyArg := make([]string, len(a))
	copy(copyArg, a)

	err := p.parse(&copyArg)
	if err != nil {
		return err
	}

	return nil
}

// in loop func
func (p *Parser) parse(args *[]string) error {
	if p.parsed {
		return nil
	}
	if len(*args) < 1 {
		return nil
	}

	err := p.parseArguemtns(args)
	if err != nil {
		return err
	}

	p.parsed = true

	if p.showHelp == true {
		p.printHelp()
	}
	return nil
}

func (p *Parser) parseArguemtns(args *[]string) error {
	n := len(p.args)

	for i := 0; i < n; i++ {
		nn := len(*args)
		oarg := p.args[i]

		for j := 0; j < nn; j++ {
			s := (*args)[j]
			if s == "" {
				continue
			}
			ct, err := oarg.check(s)
			if err != nil {
				return err
			}
			if ct > 0 {
				valpos := j + 1
				if len(*args) < valpos+oarg.size {
					return fmt.Errorf("no enough arguments for %s", oarg.name())
				}
				err := oarg.parse((*args)[valpos : valpos+oarg.size])
				if err != nil {
					return err
				}

				oarg.reduce(j, args)
				continue
			}
		}
		if oarg.opts != nil && oarg.opts.Require && !oarg.parsed {
			return fmt.Errorf("[%s] is required", oarg.name())
		}

	}

	return nil
}

func (p *Parser) String(short, long string, opts *Option) *string {
	var result string

	p.typeVar(&result, short, long, 1, true, opts)

	return &result
}
func (p *Parser) StringVar(i *string, short, long string, opts *Option) {
	p.typeVar(i, short, long, 1, true, opts)
}

func (p *Parser) Int(short, long string, opts *Option) *int {
	var result int

	p.typeVar(&result, short, long, 1, true, opts)

	return &result
}
func (p *Parser) IntVar(i *int, short, long string, opts *Option) {
	p.typeVar(i, short, long, 1, true, opts)
}

func (p *Parser) Bool(short, long string, opts *Option) *bool {
	var result bool

	p.typeVar(&result, short, long, 0, true, opts)

	return &result
}
func (p *Parser) BoolVar(i *bool, short, long string, opts *Option) {
	p.typeVar(i, short, long, 0, true, opts)
}

func (p *Parser) Float(short, long string, opts *Option) *float64 {
	var result float64

	p.typeVar(&result, short, long, 1, true, opts)

	return &result
}
func (p *Parser) FloatVar(i *float64, short, long string, opts *Option) {
	p.typeVar(i, short, long, 1, true, opts)
}

func (p *Parser) StringSlice(short, long string, opts *Option) *[]string {
	var result []string

	p.typeVar(&result, short, long, 1, false, opts)

	return &result
}
func (p *Parser) StringSliceVar(i *[]string, short, long string, opts *Option) {
	p.typeVar(i, short, long, 1, false, opts)
}

func (p *Parser) IntSlice(short, long string, opts *Option) *[]int {
	var result []int

	p.typeVar(&result, short, long, 1, false, opts)

	return &result
}
func (p *Parser) IntSliceVar(i *[]int, short, long string, opts *Option) {
	p.typeVar(i, short, long, 1, false, opts)
}

func (p *Parser) FloatSlice(short, long string, opts *Option) *[]float64 {
	var result []float64

	p.typeVar(&result, short, long, 1, false, opts)

	return &result
}
func (p *Parser) FloatSliceVar(i *[]float64, short, long string, opts *Option) {
	p.typeVar(i, short, long, 1, false, opts)
}
